@@ -45,6 +45,8 @@ extern int optind;
 
 const char *progname = "blkid";
 
+int raw_chars;
+
 static void print_version(FILE *out)
 {
 	fprintf(out, "%s from %s (libblkid %s, %s)\n",
@@ -58,35 +60,36 @@ static void usage(int error)
 	print_version(out);
 	fprintf(out,
 		"Usage:\n"
-		"  %1$s -L <label> | -U <uuid>\n\n"
-		"  %1$s [-c <file>] [-ghlLv] [-o format] [-s <tag>] \n"
-		"        [-t <token>] [-w <file>] [dev ...]\n\n"
-		"  %1$s -p [-s <tag>] [-O <offset>] [-S <size>] \n"
-		"        [-o format] <dev> [dev ...]\n\n"
-		"  %1$s -i [-s <tag>] [-o format] <dev> [dev ...]\n\n"
+		" %1$s -L <label> | -U <uuid>\n\n"
+		" %1$s [-c <file>] [-ghlLv] [-o <format>] [-s <tag>] \n"
+		"       [-t <token>] [-w <file>] [<dev> ...]\n\n"
+		" %1$s -p [-s <tag>] [-O <offset>] [-S <size>] \n"
+		"       [-o <format>] <dev> ...\n\n"
+		" %1$s -i [-s <tag>] [-o <format>] <dev> ...\n\n"
 		"Options:\n"
-		"  -c <file>   cache file (default: /etc/blkid.tab, /dev/null = none)\n"
-		"  -h          print this usage message and exit\n"
-		"  -g          garbage collect the blkid cache\n"
-		"  -o <format> output format; can be one of:\n"
-		"              value, device, list, udev, export or full; (default: full)\n"
-		"  -s <tag>    show specified tag(s) (default show all tags)\n"
-		"  -t <token>  find device with a specific token (NAME=value pair)\n"
-		"  -l          lookup the the first device with arguments specified by -t\n"
-		"  -L <label>  convert LABEL to device name\n"
-		"  -U <uuid>   convert UUID to device name\n"
-		"  -v          print version and exit\n"
-		"  -w <file>   write cache to different file (/dev/null = no write)\n"
-		"  <dev>       specify device(s) to probe (default: all devices)\n\n"
+		" -c <file>   cache file (default: /etc/blkid.tab, /dev/null = none)\n"
+		" -d          don't encode non-printing characters\n"
+		" -h          print this usage message and exit\n"
+		" -g          garbage collect the blkid cache\n"
+		" -o <format> output format; can be one of:\n"
+		"               value, device, list, udev, export or full; (default: full)\n"
+		" -k          list all known filesystems/RAIDs and exit\n"
+		" -s <tag>    show specified tag(s) (default show all tags)\n"
+		" -t <token>  find device with a specific token (NAME=value pair)\n"
+		" -l          look up only first device with token specified by -t\n"
+		" -L <label>  convert LABEL to device name\n"
+		" -U <uuid>   convert UUID to device name\n"
+		" -v          print version and exit\n"
+		" -w <file>   write cache to different file (/dev/null = no write)\n"
+		" <dev>       specify device(s) to probe (default: all devices)\n\n"
 		"Low-level probing options:\n"
-		"  -p          low-level superblocks probing (bypass cache)\n"
-		"  -i          gather information about I/O limits\n"
-		"  -S <size>   overwrite device size\n"
-		"  -O <offset> probe at the given offset\n"
-		"  -u <list>   filter by \"usage\" (e.g. -u filesystem,raid)\n"
-		"  -n <list>   filter by filesystem type (e.g. -n vfat,ext3)\n"
-		"\n",
-				progname);
+		" -p          low-level superblocks probing (bypass cache)\n"
+		" -i          gather information about I/O limits\n"
+		" -S <size>   overwrite device size\n"
+		" -O <offset> probe at the given offset\n"
+		" -u <list>   filter by \"usage\" (e.g. -u filesystem,raid)\n"
+		" -n <list>   filter by filesystem type (e.g. -n vfat,ext3)\n"
+		"\n", progname);
 
 	exit(error);
 }
@@ -104,13 +107,15 @@ static void safe_print(const char *cp, int len)
 
 	while (len--) {
 		ch = *cp++;
-		if (ch > 128) {
-			fputs("M-", stdout);
-			ch -= 128;
-		}
-		if ((ch < 32) || (ch == 0x7f)) {
-			fputc('^', stdout);
-			ch ^= 0x40; /* ^@, ^A, ^B; ^? for DEL */
+		if (!raw_chars) {
+			if (ch >= 128) {
+				fputs("M-", stdout);
+				ch -= 128;
+			}
+			if ((ch < 32) || (ch == 0x7f)) {
+				fputc('^', stdout);
+				ch ^= 0x40; /* ^@, ^A, ^B; ^? for DEL */
+			}
 		}
 		fputc(ch, stdout);
 	}
@@ -241,7 +246,7 @@ static void pretty_print_dev(blkid_dev dev)
 	pretty_print_line(devname, fs_type, label, mtpt, uuid);
 }
 
-static void print_udev_format(const char *name, const char *value, size_t sz)
+static void print_udev_format(const char *name, const char *value)
 {
 	char enc[265], safe[256];
 	size_t namelen = strlen(name);
@@ -301,7 +306,7 @@ static void print_value(int output, int num, const char *devname,
 		fputc('\n', stdout);
 
 	} else if (output & OUTPUT_UDEV_LIST) {
-		print_udev_format(name, value, valsz);
+		print_udev_format(name, value);
 
 	} else if (output & OUTPUT_EXPORT_LIST) {
 		fputs(name, stdout);
@@ -375,10 +380,9 @@ static int append_str(char **res, size_t *sz, const char *a, const char *b)
 		return -1;
 
 	str = realloc(str, len + 1);
-	if (!str) {
-		free(*res);
+	if (!str)
 		return -1;
-	}
+
 	*res = str;
 	str += *sz;
 
@@ -517,6 +521,7 @@ static int lowprobe_device(blkid_probe pr, const char *devname,
 		nvals = blkid_probe_numof_values(pr);
 
 	if (nvals &&
+	    !(chain & LOWPROBE_TOPOLOGY) &&
 	    !(output & OUTPUT_UDEV_LIST) &&
 	    !blkid_probe_has_value(pr, "TYPE") &&
 	    !blkid_probe_has_value(pr, "PTTYPE"))
@@ -615,7 +620,7 @@ static char **list_to_types(const char *list, int *flag)
 {
 	int i;
 	const char *p = list;
-	char **res;
+	char **res = NULL;
 
 	if (p && strncmp(p, "no", 2) == 0) {
 		*flag = BLKID_FLTR_NOTIN;
@@ -685,7 +690,7 @@ int main(int argc, char **argv)
 
 	show[0] = NULL;
 
-	while ((c = getopt (argc, argv, "c:f:ghilL:n:o:O:ps:S:t:u:U:w:v")) != EOF)
+	while ((c = getopt (argc, argv, "c:df:ghilL:n:ko:O:ps:S:t:u:U:w:v")) != EOF)
 		switch (c) {
 		case 'c':
 			if (optarg && !*optarg)
@@ -695,6 +700,9 @@ int main(int argc, char **argv)
 			if (!write)
 				write = read;
 			break;
+		case 'd':
+			raw_chars = 1;
+			break;
 		case 'L':
 			eval++;
 			search_value = strdup(optarg);
@@ -728,6 +736,15 @@ int main(int argc, char **argv)
 		case 'g':
 			gc = 1;
 			break;
+		case 'k':
+		{
+			size_t idx = 0;
+			const char *name = NULL;
+
+			while (blkid_superblocks_get_name(idx++, &name, NULL) == 0)
+				printf("%s\n", name);
+			exit(0);
+		}
 		case 'o':
 			if (!strcmp(optarg, "value"))
 				output_format = OUTPUT_VALUE_ONLY;
